//***************************************************************************
//
//   Initialization (by ini-file) tools v.4
//   (c) 2002-2004 Dr.Bashnyakov A. 
//   (bash@unicyb.kiev.ua)
//
//***************************************************************************
//                                                                         
//   This program is free software; you can redistribute it and/or modify  
//   it under the terms of the GNU General Public License as published by  
//   the Free Software Foundation; either version 2 of the License, or     
//   (at your option) any later version.                                   
//                                                                         
//***************************************************************************

//---------------------------------------------------------------------------
//                            About ini-file
// 
// 1. Data in ini-file are structurated by sections: section name should be 
// unique, included in "[]". All spaces around section name will be ignored.
// Remark. In simple ini-file there is no sections.
// 2. Data inside section have the NAME and the VALUE separeted by "=": 
// NAME = VALUE (extra spaces will be ignored)
// NAME - mnemonic name of a variable, VALUE - data
// Names of data should be unique inside the section and can coincide
// with names in another sections. Any alphabetical/numeric symbols and
// '_' and '$' are accepted for NAME. Values can be placed in more than one 
// strings but each string (except the last) must be terminated by '\'.
// 3. Strings whith the first non-space caracter that is not 
// alphabetical/numeric symbols nor '_', '$', '[' are comments
//
// Example
// ----------------------
// 
//    ;comment: section of Model 1
// [model]
// t0=0
// T=1
//
//    ;section of method
// [method]
// n = 2, 5, 7
// FileName = save.res, \
//            save.gif, \
//            save.bmp
// t0=0
//---------------------------------------------------------------------------


#ifndef _ini_h_
#define _ini_h_


#include <fstream> 
#include <sstream>
#include <string>


class GetDataManager;

// general method for obtain data (valueVar) from file
// by nameFile, nameSection, nameVar, separator
// return:
//  0-OK
//  1-incorrect parameters
//  2-error with opening data file
//  3-Section not found
//  4-data not found 
template <class T>
int GetData(T& valueVar, const char* nameFile, const char* nameSection, const char* nameVar, const char* separator = ",")
{
        if (!strlen(nameFile) || !strlen(nameVar)) return 1;

        bool isSimple(strlen(nameSection) == 0);
                
        std::ifstream stream(nameFile);
        if (!stream) return 2;

        std::string _nameSection(nameSection), _nameVar(nameVar), _separator("");

        GetDataManager::_StrClean(_nameSection);
        GetDataManager::_StrClean(_nameVar);

        if (separator) _separator.append(separator);

        if (GetDataManager::_GetSection(stream, _nameSection))
                return 3; 
        else if (GetDataManager::_GetEntry(valueVar, stream, _nameVar, _separator, isSimple))
                return 4; 
                
        return 0;
}

// safe method for obtain data (valueVar) from file
// by nameFile, nameSection, nameVar, separator
// if data not found default value will be used 
// return:
//  0-OK
//  1-incorrect parameters
//  2-error with opening data file
//  3-Section not found
//  4-data not found 
template <class T>
int GetDataSafe(T& valueVar, const char* nameFile, const char* nameSection, const char* nameVar, const T& dflt, const char* separator = ",")
{
        int returnValue = GetData(valueVar, nameFile, nameSection, nameVar, separator);

        if (returnValue) valueVar = dflt;
        
        return returnValue;
}

// incapsulation GetData()
class GetDataManager
{
private:
        std::string nameFile;
        std::string delim;

public:
        explicit GetDataManager(const char* name_File, const char* separ = ",") :
                nameFile(""), delim("")
        { 
                if (name_File) nameFile.append(name_File);
                if (separ) delim.append(separ);
        }

        template <class T> 
        int GetData(T& valueVar, const char* nameSection, const char* nameVar)
        {
                return ::GetData(valueVar, nameFile.c_str(), nameSection, nameVar, delim.c_str());
        }

        template <class T> 
        int GetDataSafe(T& valueVar, const char* nameSection, const char* nameVar, const T& dflt)
        {
                return ::GetDataSafe(valueVar, nameFile.c_str(), nameSection, nameVar, dflt, delim.c_str());
        }

// cleaning str: removing spaces depending mode:
// 0-from begin and from end, 1-from end only, 2-from begin only,
// str is input string and the output as well.
// return: lenght of cleaned string
        static int _StrClean(std::string &str, short mode = 0)
        {
                int i = 0, j = str.length() - 1;
                
                if (mode == 0 || mode == 2)
                        while (i <= j && isspace(str[i])) i++;
                
                if (mode == 0 || mode == 1)
                        while(j > i && isspace(str[j])) j--;
                
                if (i > j) str.erase();
                else if (i > 0 || j < str.length() - 1)
                        str.replace(0, str.length(), str.substr(i, j - i + 1)); 
                
                return str.length();
        }

// reading portion of data from file 
// return: 
//  0-OK, 
//  1-end of file or any errors occured
        static int _GetDataPortion(std::string& dest, std::ifstream& stream) 
        {
                dest.erase();

                if (stream.rdstate()) return 1;

                std::string line;
                std::getline(stream, line);

                if (_StrClean(line, 2) && (isalnum(line[0]) || line[0] == '$' || line[0] == '_') && line.find('=') != line.npos)
                {                
                        while (_StrClean(line, 1) && line[line.length() - 1] == '\\')
                        {
                                line.erase(line.length() - 1);

                                if (stream.rdstate()) break;
                                
                                dest.append(line);
                                
                                std::getline(stream, line);
                        }
                }
                
                dest.append(line);

                _StrClean(dest, 1);

                return 0;
        }

// obtaining stream that points to nameSection
// return:
//  0-OK, section is found and stream points to the first var of section
//  1-section not found
        static int _GetSection(std::ifstream& stream, const std::string& nameSection) 
        {
                if (!nameSection.length()) return 0;
                
                std::string line;

                while (!_GetDataPortion(line, stream))
                {
                        if (line.length() > 0 && line[0] == '[')
                        {
                                int closeBrace = line.find(']');

                                if (closeBrace != line.npos && closeBrace > 1) 
                                {
                                        line.replace(0, line.length(), line.substr(1, closeBrace - 1));
                                        if (_StrClean(line) && !line.compare(nameSection))
                                                return 0;
                                }
                        }
                }

                return 1;
        }

// obtaining valueVar by its nameVar in current section 
// return:
//  0: OK
//  1: data not found
        template <class T>
        static int _GetEntry(T& valueVar, std::ifstream& stream, const std::string& nameVar, const std::string& separator, bool isSimple)
        {
                int separatorPos, cnt;

                std::string line, lineVar, lineValue;

                while (!_GetDataPortion(line, stream))
                {
                        if (!isSimple && line.length() && line[0] == '[' && line.find(']') != line.npos) 
                                        return 1;  //e.g. new section began

                        if (line.find('=') != line.npos)
                        {
                                lineValue.erase();
                                lineVar.erase();
                                lineVar.append(line.substr(0,line.find('=')));

                                if (_StrClean(lineVar) && !lineVar.compare(nameVar))
                                        if(_StrClean(lineValue.append(line.substr(line.find('=') + 1))))
                                        {
                                                for (cnt = 0; cnt < separator.length(); cnt++)
                                                        while ((separatorPos = lineValue.find(separator[cnt])) != lineValue.npos)
                                                                lineValue.replace(separatorPos, 1, " ");

                                                std::istringstream strstr(lineValue.c_str());

                                                strstr >> valueVar;
                                                
                                                return 0;
                                        }
                        }
                }

                return 1;
        }
};


#endif  
